//////////////////////////////////////////////////////////////
//
//  Description	: PDF2HTML/BMP software
//  Author	: verypdf.com Inc
//  Version	: Ver1.2
//  Platform	: Windows95/98/Me/NT/2000/XP
//  Environment	: Visual C++ 6.0
//  Date	: 2002/04/03
//
//  Copyright(C) 2000--2003 verypdf.com Inc
//      
//  Contact infomation:
//        http://www.verypdf.com/
//        support@verypdf.com
//
//  This program is shareware; you can redistribute and/or 
//  modify it under the terms of the GNU General Public License.
//  See the GNU General Public License for more details.
//
//               Portion base on Xpdf code
//         Copyright 1996-2003 Glyph && Cog, LLC
//////////////////////////////////////////////////////////////

#ifdef __GNUC__
#pragma implementation
#endif

#include <aconf.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "gmem.h"
#include "gfile.h"
#include "GString.h"
#include "Error.h"
#include "GlobalParams.h"
#include "CMap.h"

//------------------------------------------------------------------------

struct CMapVectorEntry {
  GBool isVector;
  union {
    CMapVectorEntry *vector;
    CID cid;
  };
};

//------------------------------------------------------------------------

CPDFMap *CPDFMap::parse(CMapCache *cache, GString *collectionA,
		  GString *cMapNameA) {
  FILE *f;
  CPDFMap *cmap;
  char buf[256];
  GBool inCodeSpace, inCIDRange;
  char *tok1, *tok2, *tok3;
  Guint start, end;
  Guint n;

  if (!(f = globalParams->findCMapFile(collectionA, cMapNameA))) {

    // Check for an identity CPDFMap.
    if (!cMapNameA->cmp("Identity") || !cMapNameA->cmp("Identity-H")) {
      return new CPDFMap(collectionA->copy(), cMapNameA->copy(), 0);
    }
    if (!cMapNameA->cmp("Identity-V")) {
      return new CPDFMap(collectionA->copy(), cMapNameA->copy(), 1);
    }

    error(-1, "Couldn't find '%s' CMap file for '%s' collection",
	  cMapNameA->getCString(), collectionA->getCString());
    return NULL;
  }

  cmap = new CPDFMap(collectionA->copy(), cMapNameA->copy());

  inCodeSpace = inCIDRange = gFalse;
  while (getLine(buf, sizeof(buf), f)) {
    tok1 = strtok(buf, " \t\r\n");
    if (!tok1 || tok1[0] == '%') {
      continue;
    }
    tok2 = strtok(NULL, " \t\r\n");
    tok3 = strtok(NULL, " \t\r\n");
    if (inCodeSpace) {
      if (!strcmp(tok1, "endcodespacerange")) {
	inCodeSpace = gFalse;
      } else if (tok2 && tok1[0] == '<' && tok2[0] == '<' &&
		 (n = strlen(tok1)) == strlen(tok2) &&
		 n >= 4 && (n & 1) == 0) {
	tok1[n - 1] = tok2[n - 1] = '\0';
	sscanf(tok1 + 1, "%x", &start);
	sscanf(tok2 + 1, "%x", &end);
	n = (n - 2) / 2;
	cmap->addCodeSpace(cmap->vector, start, end, n);
      }
    } else if (inCIDRange) {
      if (!strcmp(tok1, "endcidrange")) {
	inCIDRange = gFalse;
      } else if (tok2 && tok3 && tok1[0] == '<' && tok2[0] == '<' &&
		 (n = strlen(tok1)) == strlen(tok2) &&
		 n >= 4 && (n & 1) == 0) {
	tok1[n - 1] = tok2[n - 1] = '\0';
	sscanf(tok1 + 1, "%x", &start);
	sscanf(tok2 + 1, "%x", &end);
	n = (n - 2) / 2;
	cmap->addCIDs(start, end, n, (CID)atoi(tok3));
      }
    } else if (tok2 && !strcmp(tok2, "usecmap")) {
      if (tok1[0] == '/') {
	cmap->useCMap(cache, tok1 + 1);
      }
    } else if (!strcmp(tok1, "/WMode")) {
      cmap->wMode = atoi(tok2);
    } else if (tok2 && !strcmp(tok2, "begincodespacerange")) {
      inCodeSpace = gTrue;
    } else if (tok2 && !strcmp(tok2, "begincidrange")) {
      inCIDRange = gTrue;
    }
  }

  fclose(f);

  return cmap;
}

CPDFMap::CPDFMap(GString *collectionA, GString *cMapNameA) {
  int i;

  collection = collectionA;
  cMapName = cMapNameA;
  wMode = 0;
  vector = (CMapVectorEntry *)gmalloc(256 * sizeof(CMapVectorEntry));
  for (i = 0; i < 256; ++i) {
    vector[i].isVector = gFalse;
    vector[i].cid = 0;
  }
  refCnt = 1;
}

CPDFMap::CPDFMap(GString *collectionA, GString *cMapNameA, int wModeA) {
  collection = collectionA;
  cMapName = cMapNameA;
  wMode = wModeA;
  vector = NULL;
  refCnt = 1;
}

void CPDFMap::useCMap(CMapCache *cache, char *useName) {
  GString *useNameStr;
  CPDFMap *subCMap;

  useNameStr = new GString(useName);
  subCMap = cache->getCMap(collection, useNameStr);
  delete useNameStr;
  if (!subCMap) {
    return;
  }
  copyVector(vector, subCMap->vector);
  subCMap->decRefCnt();
}

void CPDFMap::copyVector(CMapVectorEntry *dest, CMapVectorEntry *src) {
  int i, j;

  for (i = 0; i < 256; ++i) {
    if (src[i].isVector) {
      if (!dest[i].isVector) {
	dest[i].isVector = gTrue;
	dest[i].vector =
	  (CMapVectorEntry *)gmalloc(256 * sizeof(CMapVectorEntry));
	for (j = 0; j < 256; ++j) {
	  dest[i].vector[j].isVector = gFalse;
	  dest[i].vector[j].cid = 0;
	}
      }
      copyVector(dest[i].vector, src[i].vector);
    } else {
      if (dest[i].isVector) {
	error(-1, "Collision in usecmap");
      } else {
	dest[i].cid = src[i].cid;
      }
    }
  }
}

void CPDFMap::addCodeSpace(CMapVectorEntry *vec, Guint start, Guint end,
			Guint nBytes) {
  Guint start2, end2;
  int startByte, endByte, i, j;

  if (nBytes > 1) {
    startByte = (start >> (8 * (nBytes - 1))) & 0xff;
    endByte = (end >> (8 * (nBytes - 1))) & 0xff;
    start2 = start & ((1 << (8 * (nBytes - 1))) - 1);
    end2 = end & ((1 << (8 * (nBytes - 1))) - 1);
    for (i = startByte; i <= endByte; ++i) {
      if (!vec[i].isVector) {
	vec[i].isVector = gTrue;
	vec[i].vector =
	  (CMapVectorEntry *)gmalloc(256 * sizeof(CMapVectorEntry));
	for (j = 0; j < 256; ++j) {
	  vec[i].vector[j].isVector = gFalse;
	  vec[i].vector[j].cid = 0;
	}
      }
      addCodeSpace(vec[i].vector, start2, end2, nBytes - 1);
    }
  }
}

void CPDFMap::addCIDs(Guint start, Guint end, Guint nBytes, CID firstCID) {
  CMapVectorEntry *vec;
  CID cid;
  int byte;
  Guint i;

  vec = vector;
  for (i = nBytes - 1; i >= 1; --i) {
    byte = (start >> (8 * i)) & 0xff;
    if (!vec[byte].isVector) {
      error(-1, "Invalid CID (%*x - %*x) in CMap",
	    2*nBytes, start, 2*nBytes, end);
      return;
    }
    vec = vec[byte].vector;
  }
  cid = firstCID;
  for (byte = (int)(start & 0xff); byte <= (int)(end & 0xff); ++byte) {
    if (vec[byte].isVector) {
      error(-1, "Invalid CID (%*x - %*x) in CMap",
	    2*nBytes, start, 2*nBytes, end);
    } else {
      vec[byte].cid = cid;
    }
    ++cid;
  }
}

CPDFMap::~CPDFMap() {
  delete collection;
  delete cMapName;
  if (vector) {
    freeCMapVector(vector);
  }
}

void CPDFMap::freeCMapVector(CMapVectorEntry *vec) {
  int i;

  for (i = 0; i < 256; ++i) {
    if (vec[i].isVector) {
      freeCMapVector(vec[i].vector);
    }
  }
  gfree(vec);
}

void CPDFMap::incRefCnt() {
  ++refCnt;
}

void CPDFMap::decRefCnt() {
  if (--refCnt == 0) {
    delete this;
  }
}

GBool CPDFMap::match(GString *collectionA, GString *cMapNameA) {
  return !collection->cmp(collectionA) && !cMapName->cmp(cMapNameA);
}

CID CPDFMap::getCID(char *s, int len, int *nUsed) 
{
	CMapVectorEntry *vec;
	int n, i;
	
	if (!(vec = vector)) 
	{
		// identity CPDFMap
		*nUsed = 2;
		if (len < 2) 
		{
			return 0;
		}
		return ((s[0] & 0xff) << 8) + (s[1] & 0xff);
	}
	n = 0;
	while (1) 
	{
		if (n >= len) 
		{
			*nUsed = n;
			return 0;
		}
		i = s[n++] & 0xff;
		if (!vec[i].isVector) 
		{
			*nUsed = n;
			return vec[i].cid;
		}
		vec = vec[i].vector;
	}
}

//------------------------------------------------------------------------

CMapCache::CMapCache() {
  int i;

  for (i = 0; i < cMapCacheSize; ++i) {
    cache[i] = NULL;
  }
}

CMapCache::~CMapCache() {
  int i;

  for (i = 0; i < cMapCacheSize; ++i) {
    if (cache[i]) {
      cache[i]->decRefCnt();
    }
  }
}

CPDFMap *CMapCache::getCMap(GString *collection, GString *cMapName) {
  CPDFMap *cmap;
  int i, j;

  if (cache[0] && cache[0]->match(collection, cMapName)) {
    cache[0]->incRefCnt();
    return cache[0];
  }
  for (i = 1; i < cMapCacheSize; ++i) {
    if (cache[i] && cache[i]->match(collection, cMapName)) {
      cmap = cache[i];
      for (j = i; j >= 1; --j) {
	cache[j] = cache[j - 1];
      }
      cache[0] = cmap;
      cmap->incRefCnt();
      return cmap;
    }
  }
  if ((cmap = CPDFMap::parse(this, collection, cMapName))) {
    if (cache[cMapCacheSize - 1]) {
      cache[cMapCacheSize - 1]->decRefCnt();
    }
    for (j = cMapCacheSize - 1; j >= 1; --j) {
      cache[j] = cache[j - 1];
    }
    cache[0] = cmap;
    cmap->incRefCnt();
    return cmap;
  }
  return NULL;
}
