//////////////////////////////////////////////////////////////
//
//  Description	: PDF2HTML/BMP software
//  Author	: verypdf.com Inc
//  Version	: Ver1.2
//  Platform	: Windows95/98/Me/NT/2000/XP
//  Environment	: Visual C++ 6.0
//  Date	: 2002/04/03
//
//  Copyright(C) 2000--2003 verypdf.com Inc
//      
//  Contact infomation:
//        http://www.verypdf.com/
//        support@verypdf.com
//
//  This program is shareware; you can redistribute and/or 
//  modify it under the terms of the GNU General Public License.
//  See the GNU General Public License for more details.
//
//               Portion base on Xpdf code
//         Copyright 1996-2003 Glyph && Cog, LLC
//////////////////////////////////////////////////////////////

#ifdef __GNUC__
#pragma implementation
#endif

#include <aconf.h>
#include <stddef.h>
#include "gmem.h"
#include "Object.h"
#include "XRef.h"
#include "Array.h"
#include "Dict.h"
#include "Page.h"
#include "Error.h"
#include "Link.h"
#include "Catalog.h"

//------------------------------------------------------------------------
// Catalog
//------------------------------------------------------------------------

Catalog::Catalog(XRef *xrefA, GBool printCommands) 
{
	Object catDict, pagesDict;
	Object obj, obj2;
	int numPages0;
	int i;
	
	ok = gTrue;
	xref = xrefA;
	pages = NULL;
	pageRefs = NULL;
	numPages = pagesSize = 0;
	baseURI = NULL;
	
	xref->getCatalog(&catDict);
	if (!catDict.isDict()) 
	{
		error(-1, "Catalog object is wrong type (%s)", catDict.getTypeName());
		goto err1;
	}
	
	// read page tree
	catDict.dictLookup("Pages", &pagesDict);
	// This should really be isDict("Pages"), but I've seen at least one
	// PDF file where the /Type entry is missing.
	if (!pagesDict.isDict()) 
	{
		error(-1, "Top-level pages object is wrong type (%s)",
			pagesDict.getTypeName());
		goto err2;
	}
	pagesDict.dictLookup("Count", &obj);
	if (!obj.isInt()) 
	{
		error(-1, "Page count in top-level pages object is wrong type (%s)",
			obj.getTypeName());
		goto err3;
	}
	pagesSize = numPages0 = obj.getInt();
	obj.free();
	pages = (Page **)gmalloc(pagesSize * sizeof(Page *));
	pageRefs = (Ref *)gmalloc(pagesSize * sizeof(Ref));
	for (i = 0; i < pagesSize; ++i) 
	{
		pages[i] = NULL;
		pageRefs[i].num = -1;
		pageRefs[i].gen = -1;
	}
	numPages = readPageTree(pagesDict.getDict(), NULL, 0, printCommands);
	if (numPages != numPages0) 
	{
		error(-1, "Page count in top-level pages object is incorrect");
	}
	pagesDict.free();
	
	// read named destination dictionary
	catDict.dictLookup("Dests", &dests);
	
	// read root of named destination tree
	if (catDict.dictLookup("Names", &obj)->isDict())
		obj.dictLookup("Dests", &nameTree);
	else
		nameTree.initNull();
	obj.free();
	
	// read base URI
	if (catDict.dictLookup("URI", &obj)->isDict()) 
	{
		if (obj.dictLookup("Base", &obj2)->isString()) 
		{
			baseURI = obj2.getString()->copy();
		}
		obj2.free();
	}
	obj.free();
	
	// get the metadata stream
	catDict.dictLookup("Metadata", &metadata);
	
	// get the structure tree root
	catDict.dictLookup("StructTreeRoot", &structTreeRoot);

	catDict.free();
	return;
	
err3:
	obj.free();
err2:
	pagesDict.free();
err1:
	catDict.free();
	dests.initNull();
	nameTree.initNull();
	ok = gFalse;
}

Catalog::~Catalog() 
{
	int i;
	
	if (pages) 
	{
		for (i = 0; i < pagesSize; ++i) 
		{
			if (pages[i]) {
				delete pages[i];
			}
		}
		gfree(pages);
		gfree(pageRefs);
	}
	dests.free();
	nameTree.free();
	if (baseURI) {
		delete baseURI;
	}
	metadata.free();
	structTreeRoot.free();
}
//////////////////////////////////////////////////////////////
//  getBookmark  ȡǩ
//    : szFileName -- ǩļƴŵ·
//          size -- szFileName length
//  Author: xueheng
//  Date  : 2002/07/10
//////////////////////////////////////////////////////////////
int Catalog::getBookmark(char* szFileName,int size)
{
	if(ok == gFalse || xref == NULL || szFileName == NULL || size <= 0)
		return gFalse;
	//get Outlines from catalog dictionary
	Object catDict;
	Object obj, obj2;

	xref->getCatalog(&catDict);
	if (!catDict.isDict()) 
	{
		catDict.free();
		return gFalse;
	}

	int outlineCount = 0;
	{
		char path_buffer[_MAX_PATH];
		char drive[_MAX_DRIVE];
		char dir[_MAX_DIR];
		char fname[_MAX_FNAME];
		char ext[_MAX_EXT];
		GetModuleFileName(NULL,path_buffer,_MAX_PATH);
		_splitpath(path_buffer, drive, dir, fname, ext );
		_makepath( path_buffer, drive, dir,"outlines",".txt");
		unlink(path_buffer);

		Object outlinesDict;
		if (catDict.dictLookup("Outlines", &outlinesDict)->isDict())
		{
			outlinesDict.dictLookup("Count", &obj);
			if (obj.isInt())
				outlineCount = obj.getInt();
			obj.free();
			if(outlineCount > 0)
			{
				nesting = -1;
				FILE* file = fopen(path_buffer,"w");
				readOutlinesTree(outlinesDict,file);
				fclose(file);
			}
		}
		outlinesDict.free();
		if(strlen(path_buffer) < size)
			strcpy(szFileName,path_buffer);
		else
			strncpy(szFileName,path_buffer,size);
	}
	catDict.free();
	if(outlineCount > 0)
		return gTrue;
	return gFalse;
}
//////////////////////////////////////////////////////////////
//
//  readOutlinesTreeȡĿ¼
//  outlinesDict
//        Outline's Root(ٸ)
//  Author: xueheng
//  Date  : 2002/04/21
//////////////////////////////////////////////////////////////
int Catalog::readOutlinesTree(Object& outlinesDict,FILE* file)
{
	Object firstOutlinesDist,lastOutlinesDist,obj;
	
	outlinesDict.dictLookup("First", &firstOutlinesDist);
	outlinesDict.dictLookup("Last", &lastOutlinesDist);
	if(!firstOutlinesDist.isDict() || !lastOutlinesDist.isDict())
	{
		firstOutlinesDist.free();
		lastOutlinesDist.free();
		return 1;
	}
	//¼Ƕ׵Ĳ
	nesting++;
	do{
		//Title
		firstOutlinesDist.dictLookup("Title", &obj);
		if (obj.isString() )
		{
			if(file)
			{
				for(int i = 0; i < nesting; i++)
					fputc('\t',file);
				fprintf(file,"%s",obj.getString()->getCString());
			}
		}
		obj.free();
		//Dest
		firstOutlinesDist.dictLookup("Dest", &obj);
		{
			LinkDest *dest = NULL;
			int page = 0;
			if (obj.isString() )
			{//Destͣ/Dest (G3.127)
				dest = findDest(obj.getString());
				//TRACE("\t%s",obj.getString()->getCString());
			}
			else if (obj.isArray() )
			{//Destͣ/Dest[8 0 R/XYZ null null 0]
				dest = new LinkDest(obj.getArray(), gTrue);
			}
			else if (obj.isDict()) 
			{
				Object obj2;
				if (obj.dictLookup("D", &obj2)->isArray())
					dest = new LinkDest(obj2.getArray(), gTrue);
				else
					error(-1, "Bad named destination value");
				obj2.free();
			}
			if(obj.isNull())
			{
				Object obj2,obj3;
				if (firstOutlinesDist.dictLookup("A", &obj2)->isDict())
				{
					if (obj2.dictLookup("D", &obj3)->isArray())
						dest = new LinkDest(obj3.getArray(), gTrue);
				}
				obj2.free();
				obj3.free();
			}
			else
			{
				error(-1, "Bad named destination value");
			}
			if (dest)
			{ 
				if (dest->isPageRef())
				{
					Ref pageref = dest->getPageRef();
					page = findPage(pageref.num,pageref.gen);
				}
				else  
					page = dest->getPageNum();
				if(file)
				{
					fprintf(file,"<>,%d,%d,%d,%d,%d,%d,%d,%d,%d\n",
						page,dest->getLeft(),dest->getTop(),
						dest->getRight(),dest->getBottom(),
						dest->getZoom(),dest->getChangeLeft(),
						dest->getChangeTop(),
						dest->getChangeZoom()
						);
				}
				delete dest;
			}
			else
			{
				if(file)
					fputc('\n',file);
			}
		}
		obj.free();
		//ǶײѯOutlinesbegin
		{
			int dictLength = firstOutlinesDist.dictGetLength();
			BOOL firstIn = FALSE,lastIn = FALSE;
			for(int i = 0; i < dictLength; i++)
			{
				if(!firstIn && strcmp("First",firstOutlinesDist.dictGetKey(i)) == 0)
					firstIn = TRUE;
				if(!lastIn && strcmp("Last",firstOutlinesDist.dictGetKey(i)) == 0)
					lastIn = TRUE;
				if(firstIn && lastIn)
					break;
			}
			if(firstIn && lastIn)
				readOutlinesTree(firstOutlinesDist,file);
		}
		//ǶײѯOutlinesend
		firstOutlinesDist.dictLookup("Next", &obj);
		if(!obj.isDict())
		{
			obj.free();
			break;
		}
		firstOutlinesDist.free();
		obj.copy(&firstOutlinesDist);
		obj.free();
	}while(firstOutlinesDist.isDict());

	//¼Ƕ׵Ĳ
	nesting--;

	firstOutlinesDist.free();
	lastOutlinesDist.free();
	return 0;
}
GString *Catalog::readMetadata() 
{
	GString *s;
	Dict *dict;
	Object obj;
	int c;
	
	if (!metadata.isStream()) {
		return NULL;
	}
	dict = metadata.streamGetDict();
	if (!dict->lookup("Subtype", &obj)->isName("XML")) {
		error(-1, "Unknown Metadata type: '%s'\n",
			obj.isName() ? obj.getName() : "???");
	}
	obj.free();
	s = new GString();
	metadata.streamReset();
	while ((c = metadata.streamGetChar()) != EOF) {
		s->append(c);
	}
	metadata.streamClose();
	return s;
}

int Catalog::readPageTree(Dict *pagesDict, PageAttrs *attrs, int start,
			  GBool printCommands) 
{
	Object kids;
	Object kid;
	Object kidRef;
	PageAttrs *attrs1, *attrs2;
	Page *page;
	int i, j;
	
	attrs1 = new PageAttrs(attrs, pagesDict);
	pagesDict->lookup("Kids", &kids);
	if (!kids.isArray()) 
	{
		error(-1, "Kids object (page %d) is wrong type (%s)",
			start+1, kids.getTypeName());
		goto err1;
	}
	for (i = 0; i < kids.arrayGetLength(); ++i) 
	{
		kids.arrayGet(i, &kid);
		if (kid.isDict("Page")) 
		{
			attrs2 = new PageAttrs(attrs1, kid.getDict());
			page = new Page(xref, start+1, kid.getDict(), attrs2, printCommands);
			if (!page->isOk()) 
			{
				++start;
				goto err3;
			}
			if (start >= pagesSize) 
			{
				pagesSize += 32;
				pages = (Page **)grealloc(pages, pagesSize * sizeof(Page *));
				pageRefs = (Ref *)grealloc(pageRefs, pagesSize * sizeof(Ref));
				for (j = pagesSize - 32; j < pagesSize; ++j) 
				{
					pages[j] = NULL;
					pageRefs[j].num = -1;
					pageRefs[j].gen = -1;
				}
			}
			pages[start] = page;
			kids.arrayGetNF(i, &kidRef);
			if (kidRef.isRef()) 
			{
				pageRefs[start].num = kidRef.getRefNum();
				pageRefs[start].gen = kidRef.getRefGen();
			}
			kidRef.free();
			++start;
			// This should really be isDict("Pages"), but I've seen at least one
			// PDF file where the /Type entry is missing.
		} 
		else if (kid.isDict()) 
		{
			if ((start = readPageTree(kid.getDict(), attrs1, start, printCommands)) < 0)
				goto err2;
		} 
		else 
		{
			error(-1, "Kid object (page %d) is wrong type (%s)",
				start+1, kid.getTypeName());
			goto err2;
		}
		kid.free();
	}
	delete attrs1;
	kids.free();
	return start;
	
err3:
	delete page;
err2:
	kid.free();
err1:
	kids.free();
	delete attrs1;
	ok = gFalse;
	return -1;
}

int Catalog::findPage(int num, int gen) 
{
	int i;
	
	for (i = 0; i < numPages; ++i) 
	{
		if (pageRefs[i].num == num && pageRefs[i].gen == gen)
			return i + 1;
	}
	return 0;
}

LinkDest *Catalog::findDest(GString *name) 
{
	LinkDest *dest;
	Object obj1, obj2;
	GBool found;
	
	// try named destination dictionary then name tree
	found = gFalse;
	if (dests.isDict()) 
	{
		if (!dests.dictLookup(name->getCString(), &obj1)->isNull())
			found = gTrue;
		else
			obj1.free();
	}
	if (!found && nameTree.isDict()) 
	{
		if (!findDestInTree(&nameTree, name, &obj1)->isNull())
			found = gTrue;
		else
			obj1.free();
	}
	if (!found)
		return NULL;
	
	// construct LinkDest
	dest = NULL;
	if (obj1.isArray()) 
	{
		dest = new LinkDest(obj1.getArray(), gTrue);
	}
	else if (obj1.isDict()) 
	{
		if (obj1.dictLookup("D", &obj2)->isArray())
			dest = new LinkDest(obj2.getArray(), gTrue);
		else
			error(-1, "Bad named destination value");
		obj2.free();
	}
	else 
	{
		error(-1, "Bad named destination value");
	}
	obj1.free();
	
	return dest;
}

Object *Catalog::findDestInTree(Object *tree, GString *name, Object *obj) 
{
	Object names, name1;
	Object kids, kid, limits, low, high;
	GBool done, found;
	int cmp, i;
	
	// leaf node
	if (tree->dictLookup("Names", &names)->isArray()) 
	{
		done = found = gFalse;
		for (i = 0; !done && i < names.arrayGetLength(); i += 2) 
		{
			if (names.arrayGet(i, &name1)->isString()) 
			{
				cmp = name->cmp(name1.getString());
				if (cmp == 0) 
				{
					names.arrayGet(i+1, obj);
					found = gTrue;
					done = gTrue;
				}
				else if (cmp < 0) 
				{
					done = gTrue;
				}
				name1.free();
			}
		}
		names.free();
		if (!found)
			obj->initNull();
		return obj;
	}
	names.free();
	
	// root or intermediate node
	done = gFalse;
	if (tree->dictLookup("Kids", &kids)->isArray()) 
	{
		for (i = 0; !done && i < kids.arrayGetLength(); ++i) 
		{
			if (kids.arrayGet(i, &kid)->isDict()) 
			{
				if (kid.dictLookup("Limits", &limits)->isArray()) 
				{
					if (limits.arrayGet(0, &low)->isString() &&
						name->cmp(low.getString()) >= 0) 
					{
						if (limits.arrayGet(1, &high)->isString() &&
							name->cmp(high.getString()) <= 0) 
						{
							findDestInTree(&kid, name, obj);
							done = gTrue;
						}
						high.free();
					}
					low.free();
				}
				limits.free();
			}
			kid.free();
		}
	}
	kids.free();
	
	// name was outside of ranges of all kids
	if (!done)
		obj->initNull();
	
	return obj;
}
