// MyMenu.cpp: implementation of the CMyMenu class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "MyMenu.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

CMyMenu *pMenu;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
IMPLEMENT_DYNAMIC(CMyMenu, CMenu);
///////////////////////////////////////////////////////////////////////////
//
// CMyMenu

CMyMenu::CMyMenu()
{
    m_penBack.CreatePen(PS_SOLID, 0, GetSysColor(COLOR_MENU));

    ZeroMemory ((PVOID) &m_lf,sizeof (LOGFONT));

    NONCLIENTMETRICS nm;

    nm.cbSize = sizeof (NONCLIENTMETRICS);

	VERIFY (SystemParametersInfo (SPI_GETNONCLIENTMETRICS,0,&nm,0)); 

    m_lf =  nm.lfMenuFont;
    m_fontMenu.CreateFontIndirect (&m_lf);

	nIconWidth = 16;
	nIconHeight = 16;

	hColorWhite   = RGB(255, 255, 255);
	hColorText    = GetSysColor(COLOR_MENUTEXT);
	hColorBlack   = RGB(0, 0, 0);
	hColorMenu	  = RGB(107, 162, 214);
	hColorGray    = GetSysColor(COLOR_GRAYTEXT);
	hColorMenuHL  = RGB(57,109,165);
	hColorMenuFHL = RGB(8,36, 107);
	hColorClear   = GetSysColor(COLOR_MENU);

	hMenuParent = NULL;

	bPopped = false;
}

CMyMenu::~CMyMenu()
{
	m_MenuShadow.RemoveShadow();

   if((HFONT)m_fontMenu != NULL)
        m_fontMenu.DeleteObject();

	for(int i = 0; i != m_MenuArray.GetSize(); i++)
		delete m_MenuArray[i];
}
	
///////////////////////////////////////////////////////////////////////////
//
// CMyMenu message handlers

VOID CMyMenu::WaitForMenu(HWND hWnd, UINT message, UINT iTimerID, DWORD dwTimer)
{
	TRACE("In Timer...\n");

	// redraws the whole shadow once all menu children have been
	// closed

	for(int i = 0; i != pMenu->hMenuChild.GetSize(); i++)
		if(pMenu->hMenuChild[i]->m_MenuShadow.IsDrawn())
			return;

	pMenu->m_MenuShadow.RedrawOldShadow();

	HWND hAfxWnd = AfxGetApp()->GetMainWnd()->GetSafeHwnd();

	::KillTimer(hAfxWnd, 41014);
}

void CMyMenu::DrawItem(LPDRAWITEMSTRUCT lpDIS)
{
	ASSERT(lpDIS != NULL);

	MENUDATA* mitem = (MENUDATA*)lpDIS->itemData;

	CRect rect;
	
	rect.CopyRect(&lpDIS->rcItem);

	CDC* pDC = CDC::FromHandle(lpDIS->hDC);

	if(!pDC) return;

	HICON hDrawIcon = mitem->hIconNormal;

	COLORREF hDrawTextColor = hColorText;

	UINT action = lpDIS->itemAction;
	UINT state  = lpDIS->itemState;

	CString strText = mitem->szName;

	LOGFONT lf = m_lf;

	CFont dispFont;
	CFont *pFont;

	CBrush pBrush;

	// draw the shadow

	if(hMenuParent != NULL)
	{
		if(!m_MenuShadow.IsDrawn())
		{
			// if two popups are next to each other and one is displayed
			// after the other, the popups' parent menu's 
			// shadow needs to be redrawn with the exception of the area
			// of the popup that is now open.

			CRect rOld;
			CWnd *pWnd = pDC->GetWindow();

			if (pWnd)
			{
				pWnd->GetWindowRect(rOld);
				hMenuParent->m_MenuShadow.RedrawOldShadowIgnore(rOld);
			}

			m_MenuShadow.DrawShadow(pDC);
		}

		if(mitem->nFlags & MF_POPUP)
		{
			// wait until menu is closed and then
			// redraw the old shadow

			if(bPopped)
			{
				pMenu = this;

				HWND hAfxWnd = AfxGetApp()->GetMainWnd()->GetSafeHwnd();
				AfxGetApp()->
				::SetTimer(hAfxWnd, 41014, 100, (TIMERPROC)WaitForMenu);

				bPopped = false;
			}	

			if(state & ODS_SELECTED)
				bPopped = true;	
		}
	}

	TRACE("Drawing... " + strText + "\n");

	// now draw the menu item

	if((mitem->nFlags & MF_SEPARATOR) || (mitem->nFlags & MF_VERTSEPARATOR))
	{
		// draw horizontal and vertical separators

		CRect rect1 = rect;

		if(mitem->nFlags & MF_VERTSEPARATOR)
		{
			rect1.left += 3;
			rect1.right = rect1.left +=2;
			rect1.bottom = rect1.top + rect1.Height();
			rect1.top += 3;
		}
		else
		{
			rect1.top += 3;
			rect1.right -= 4;
			rect1.left += 4;
			rect1.bottom = rect1.top + 2;
		}

		pDC->Draw3dRect(rect1, hColorGray, hColorWhite);

		hDrawIcon = NULL;
	}
	else if((state & ODS_SELECTED) || (state & ODS_HOTLIGHT))
	{
		// draw selected state

		if(state & ODS_DISABLED && !(state & ODS_HOTLIGHT))
		{
			pBrush.CreateSolidBrush(hColorGray + RGB(70, 70, 70));

			pDC->FillRect(rect, &pBrush);

			if(!(state & ODS_DISABLED))
				lf.lfWeight = FW_BOLD;

			hDrawTextColor = hColorWhite;

			pBrush.DeleteObject();
		}
		else if((mitem->nFlags & MF_BEVELED) || (!hMenuParent && !(mitem->nFlags & MF_FLAT)) )
		{
			pBrush.CreateSolidBrush(hColorClear);
			
			pDC->FillRect(rect, &pBrush);

			CRect rect1 = rect;

			rect1.bottom--;
			rect1.right--;

			pBrush.DeleteObject();
			pBrush.CreateSolidBrush(hColorMenu);

			pDC->FillRect(rect1, &pBrush);

			rect1.bottom ++;
			rect1.right++;

			pDC->Draw3dRect(rect1, hColorMenu, hColorMenuHL);

			rect1.top++;
			rect1.left++;

			pDC->Draw3dRect(rect1, hColorWhite, hColorGray - RGB(50, 50, 50));

			pBrush.DeleteObject();
		}
		else 
		{
			// draw non-selected state

			pBrush.CreateSolidBrush(hColorMenu);

			pDC->FillRect(rect, &pBrush);

			pDC->Draw3dRect(rect, hColorMenuFHL, hColorMenuFHL);

			pBrush.DeleteObject();
		}

		if(!(state & ODS_DISABLED))
			lf.lfWeight = FW_BOLD;

		hDrawTextColor = hColorWhite;
	}
	else
	{
		pBrush.CreateSolidBrush(hColorClear);

		rect.bottom++;
		rect.right++;

		pDC->FillRect(rect, &pBrush);

		if(mitem->nFlags & MF_DEFAULT)
			lf.lfWeight = FW_BOLD;

		pBrush.DeleteObject();
	}

	if((HFONT)dispFont != NULL)

		dispFont.DeleteObject();

	dispFont.CreateFontIndirect(&lf);

	rect.left += 4;
	rect.top++;

	// draw icon

	if(hDrawIcon != NULL || state & ODS_CHECKED)
	{
		CPoint pt;
		CSize size;
		
		pt.x = rect.left;
		pt.y = rect.top + 3;

		size.cx = nIconWidth;
		size.cy = nIconHeight;

		pBrush.CreateSolidBrush(hColorMenuFHL + RGB(70, 70, 70));
		
		CRect rect1 = rect;

		rect1.left += 2;
		rect1.top += 4;
		rect1.right = rect1.left + nIconWidth - 3;
		rect1.bottom = rect1.top + nIconHeight - 3;

		if(state & ODS_SELECTED)
		{
			pDC->DrawState(pt, size, hDrawIcon, DSS_NORMAL | DSS_MONO, pBrush);
			pt.x--; pt.y--;
		}

		if(state & ODS_CHECKED)
			pDC->DrawFrameControl(rect1, DFC_BUTTON, DFCS_CHECKED | DFCS_MONO);
		else if(state & ODS_DISABLED)
			pDC->DrawState(pt, size, hDrawIcon, DSS_DISABLED, pBrush);
		else
			pDC->DrawState(pt, size, hDrawIcon, DSS_NORMAL, pBrush);

		pBrush.DeleteObject();
	}

	// compensate for icon when drawing text

	if(hMenuParent == NULL && !(mitem->nFlags & MF_PRETENDICON))
	{
		if(mitem->hIconNormal != NULL)
			rect.left += nIconWidth;
	}
	else
		rect.left += nIconWidth;

	rect.left += 5;

	// draw text

	if(!(mitem->nFlags & MF_VERTSEPARATOR) && !(mitem->nFlags & MF_SEPARATOR))
	if(!strText.IsEmpty())
	{
		int	iOldMode = pDC->GetBkMode();
		
		pDC->SetBkMode(TRANSPARENT);

		CString strAccel = "";

		if(strText.Find("\t") != -1)
		{
			strAccel = strText.Mid(strText.Find("\t") + 1, strText.GetLength());
			strText = strText.Mid(0, strText.Find("\t"));
		}

		if((state & ODS_DISABLED) && !(state & ODS_SELECTED))
		{
			rect.left++;
			rect.top++;

			pDC->SetTextColor(hColorWhite);

			pFont = pDC->SelectObject (&dispFont);
			pDC->DrawText(strText, rect, DT_LEFT | DT_SINGLELINE | DT_VCENTER);

			pDC->SetTextColor(hColorGray);

			rect.left--;
			rect.top--;
		}
		else if((state & ODS_SELECTED) && (state & ODS_DISABLED))
			pDC->SetTextColor(hColorWhite);
		else
			pDC->SetTextColor(hDrawTextColor);
		
		pFont = pDC->SelectObject (&dispFont);
		
		if(hMenuParent == NULL && !(state & ODS_DISABLED))
			rect.top -= 2;

		pDC->DrawText(strText, rect, DT_LEFT | DT_SINGLELINE | DT_VCENTER);

		if(strAccel != "")
		{
			strAccel.TrimRight();
			strAccel += "    ";

			pDC->DrawText(strAccel, rect, DT_RIGHT | DT_SINGLELINE | DT_VCENTER);
		}

		pDC->SetBkMode(iOldMode);

		pDC->SelectObject(pFont); 
	}

	dispFont.DeleteObject();
}

void CMyMenu::MeasureItem(LPMEASUREITEMSTRUCT lpMIS)
{
	CWnd  *pWnd  = AfxGetMainWnd();		
    CDC   *pDC   = pWnd->GetDC();				
    CFont *pFont = pDC->SelectObject(&m_fontMenu);

	MENUDATA* mitem = (MENUDATA*)lpMIS->itemData;

	SIZE n;

	GetTextExtentPoint32(pDC->GetSafeHdc(), mitem->szName, lstrlen(mitem->szName), &n); 

    pDC->SelectObject(pFont);
	
    AfxGetApp()->m_pMainWnd->ReleaseDC(pDC);

	lpMIS->itemHeight = nIconHeight + 6;
	lpMIS->itemWidth = nIconWidth + 20 + n.cx;

	if(hMenuParent == NULL && !(mitem->nFlags & MF_PRETENDICON))
		if(mitem->hIconNormal == NULL)
			lpMIS->itemWidth = n.cx;

	if(strstr(mitem->szName, "\t") != NULL)
		lpMIS->itemWidth += 15;

	if(mitem->nFlags & MF_SEPARATOR)

		lpMIS->itemHeight = 7;

	if(mitem->nFlags & MF_VERTSEPARATOR)
	{
		lpMIS->itemHeight = nIconHeight + 3;
		lpMIS->itemWidth = 0;
	}
}

void CMyMenu::SetParent(CMyMenu *hMenu)
{
	hMenuParent = hMenu;

	hMenu->hMenuChild.Add(this);
}

HICON CMyMenu::LoadIcon(UINT nResource)
{
	HINSTANCE hAfx = AfxGetInstanceHandle();

	return (HICON)::LoadImage(hAfx, (LPCTSTR)nResource, IMAGE_ICON, 0,0, 
								LR_CREATEDIBSECTION);
}

void CMyMenu::AppendMenu(char* szName, UINT nID, UINT nFlags, HICON hIconNormal)
{
	MENUDATA* mitem = new MENUDATA;

	mitem->nFlags = nFlags | MF_OWNERDRAW;
	mitem->nID = nID;
	mitem->hIconNormal = hIconNormal;

	lstrcpy(mitem->szName, szName);

	int n = m_MenuArray.Add(mitem);

	CMenu::AppendMenu(mitem->nFlags, mitem->nID, (LPCTSTR)mitem);
}

//*************************************************************************



