// Shadow.cpp: implementation of the CShadow class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "win_pdf2rtf.h"
#include "Shadow.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CShadow::CShadow()
{
	nIntensity = 50;
	nBorder = 5;

	bDrawn = false;
	
	pDestDC = CDC::FromHandle(::CreateDC("DISPLAY", NULL, NULL, NULL));
	
	// CDC::FromHandle(::GetDC(NULL)) doesn't work on Win98SE

	pSaveWnd = NULL;
	rcSave.SetRectEmpty();

	int width = pDestDC->GetDeviceCaps(HORZRES);
	int height = pDestDC->GetDeviceCaps(VERTRES);
	
	m_Bitmap[0].CreateCompatibleBitmap(pDestDC, width, height);
	m_Bitmap[1].CreateCompatibleBitmap(pDestDC, width, height);
	m_Bitmap[2].CreateCompatibleBitmap(pDestDC, width, height);
	m_Bitmap[3].CreateCompatibleBitmap(pDestDC, width, height);

	hRightDC.CreateCompatibleDC(NULL);
	hBottomDC.CreateCompatibleDC(NULL);

	hShadowRightDC.CreateCompatibleDC(NULL);
	hShadowBottomDC.CreateCompatibleDC(NULL);

	m_OldBitmap[0] = hRightDC.SelectObject(&m_Bitmap[0]);
	m_OldBitmap[1] = hBottomDC.SelectObject(&m_Bitmap[1]);
	m_OldBitmap[2] = hShadowRightDC.SelectObject(&m_Bitmap[2]);
	m_OldBitmap[3] = hShadowBottomDC.SelectObject(&m_Bitmap[3]);
}

CShadow::~CShadow()
{
	hRightDC.SelectObject(m_OldBitmap[0]);
	hBottomDC.SelectObject(m_OldBitmap[1]);

	hShadowRightDC.SelectObject(m_OldBitmap[2]);
	hShadowBottomDC.SelectObject(m_OldBitmap[3]);

	hRightDC.DeleteDC();
	hBottomDC.DeleteDC();

	hShadowRightDC.DeleteDC();
	hShadowBottomDC.DeleteDC();

	m_Bitmap[0].DeleteObject();
	m_Bitmap[1].DeleteObject();
	m_Bitmap[2].DeleteObject();
	m_Bitmap[3].DeleteObject();
}

//////////////////////////////////////////////////////////////////////
// REMOVESHADOW ()
//////////////////////////////////////////////////////////////////////
//
//	REMOVES THE SHADOW
//
//////////////////////////////////////////////////////////////////////

void CShadow::RemoveShadow()
{
	if(bDrawn && pDestDC != NULL)
	{
		CRect r, rc;

		if(pSaveWnd != NULL)
		{
			pSaveWnd->GetWindowRect(r);
			rc = r;
		}
		else if(rcSave.IsRectEmpty())
		{
			rc = rcSave;
		}

		// remove bottom rect and then right rect, good idea
		// to always remove in this order, had problem earlier when
		// tried to do it in the other order

		r = GetBottomRect(rc);

		pDestDC->BitBlt(r.left, r.top, r.Width(), r.Height(), &hBottomDC, 0, 0, SRCCOPY);

		r = GetRightRect(rc);

		pDestDC->BitBlt(r.left, r.top, r.Width(), r.Height(), &hRightDC, 0, 0, SRCCOPY);
	}

	bDrawn = false;
}

//////////////////////////////////////////////////////////////////////
// REDRAWOLDSHADOW ()
//////////////////////////////////////////////////////////////////////
//
//	REPAINTS THE SHADOW PREVIOUSLY DRAWN
//
//////////////////////////////////////////////////////////////////////

void CShadow::RedrawOldShadow()
{
	if(bDrawn && pDestDC != NULL)
	{
		CRect r,rc;

		if(pSaveWnd != NULL)
		{
			pSaveWnd->GetWindowRect(r);
			rc = r;
		}
		else if(rcSave.IsRectEmpty())
			rc = rcSave;
		else
			return;

		if(rc.IsRectEmpty())
			return;

		r = GetBottomRect(rc);

		pDestDC->BitBlt(r.left, r.top, r.Width(), r.Height(), &hShadowBottomDC, 0, 0, SRCCOPY);

		r = GetRightRect(rc);

		pDestDC->BitBlt(r.left, r.top, r.Width(), r.Height() + nBorder, &hShadowRightDC, 0, 0, SRCCOPY);
	}
}

//////////////////////////////////////////////////////////////////////
// REDRAWOLDSHADOWIGNORE (CRECT RIGNORE)
//////////////////////////////////////////////////////////////////////
//
//	REPAINTS THE SHADOW PREVIOUSLY DRAWN WITH THE EXCEPTION OF A
//  PERTICULAR AREA
//
//////////////////////////////////////////////////////////////////////

void CShadow::RedrawOldShadowIgnore(CRect rIgnore)
{
	// RIGNORE = DO NOT DRAW RIGHT SHADOW IN THIS AREA

	if(bDrawn && pDestDC != NULL)
	{
		CRect r, rc;

		if(pSaveWnd != NULL)
		{
			pSaveWnd->GetWindowRect(r);
			rc = r;
		}
		else if(rcSave.IsRectEmpty())
			rc = rcSave;
		else
			return;

		if(rc.IsRectEmpty())
			return;

		r = GetBottomRect(rc);
		
		if(rIgnore.left <= r.right && rIgnore.bottom > r.bottom)
		{
			pDestDC->BitBlt(r.left, r.top,  rIgnore.left - r.left - 2, r.Height(), &hShadowBottomDC, 0, 0, SRCCOPY);
		}
		else
			pDestDC->BitBlt(r.left, r.top, r.Width(), r.Height(), &hShadowBottomDC, 0, 0, SRCCOPY);

		r = GetRightRect(rc);

		if(rIgnore.top <= r.bottom && (rIgnore.left <= r.left && rIgnore.right >= rIgnore.right))
		{
			int toph = rIgnore.top - r.top;

			if(rIgnore.bottom > r.bottom)
			{
				pDestDC->BitBlt(r.left, r.top, r.Width(), toph, &hShadowRightDC, 0, 0, SRCCOPY);
			}
			else
			{
				int topb = rIgnore.Height() + toph;

				pDestDC->BitBlt(r.left, r.top, r.Width(), rIgnore.top - r.top, &hShadowRightDC, 0, 0, SRCCOPY);
				pDestDC->BitBlt(r.left, r.top + topb, r.Width(), r.Height() - topb, &hShadowRightDC, 0, rIgnore.Height() + toph, SRCCOPY);
			}
		}
		else
			pDestDC->BitBlt(r.left, r.top, r.Width(), r.Height() + nBorder, &hShadowRightDC, 0, 0, SRCCOPY);
	}
}

//////////////////////////////////////////////////////////////////////
// REDRAWSHADOW ()
//////////////////////////////////////////////////////////////////////
//
//	COMPLETELY REDRAWS A SHADOW PREVIOUSLY DRAWN
//
//////////////////////////////////////////////////////////////////////

bool CShadow::RedrawShadow()
{
	if(pSaveWnd != NULL && IsDrawn())
		return DrawShadow(pSaveWnd);

	if(!rcSave.IsRectEmpty())
		return DrawShadow(rcSave);

	return false;
}

//////////////////////////////////////////////////////////////////////
// UPDATESHADOW(CDC *PDC)
//////////////////////////////////////////////////////////////////////
//
//	UPDATES INFORMATION THAT IS SAVED FOR LATER OPERATIONS
//
//////////////////////////////////////////////////////////////////////

CRect CShadow::UpdateShadow(CDC *pDC)
{
	if(!pDC) return NULL;

	pSaveWnd = pDC->GetWindow();
	pSaveWnd->GetWindowRect(rcSave);

	return rcSave;
}

//////////////////////////////////////////////////////////////////////
// DRAWSHADOW (CDC* PWND) & DRAWSHADOW (CDC* PDC) 
//////////////////////////////////////////////////////////////////////
//
//	DRAWS SHADOW USING A CWND OR PDC TO FIGURE OUT WHERE TO 
//  DRAW THE SHADOWS	
//
//////////////////////////////////////////////////////////////////////

bool CShadow::DrawShadow(CDC *pDC)
{
	if (!pDC)
		return false;

	return DrawShadow(pDC->GetWindow());
}

bool CShadow::DrawShadow(CWnd *pWnd)
{
	if (!pWnd)
		return false;

	return DrawShadow(UpdateShadow(pWnd->GetDC()));
}

//////////////////////////////////////////////////////////////////////
// DRAWSHADOW (CWND* PWND)
//////////////////////////////////////////////////////////////////////
//
//	DRAWS SHADOW USING A RECTANGLE
//
//////////////////////////////////////////////////////////////////////

bool CShadow::DrawShadow(CRect r)
{
	// we use hShadowRightDC and hShadowBottomDC instead of one DC 
	// for the menu-example purposes, when one menu exits the its submenu's
	// shadows have to be redrawn, and so we save those sections

	if(r.IsRectEmpty() || !pDestDC || !hRightDC || !hBottomDC ||
		!hShadowRightDC || !hShadowBottomDC)
		return false;

	int i, y, x, c, in = 0;

	// draw the right shadow

	rcSave = r;

	r = GetRightRect(rcSave);

	hRightDC.BitBlt(0, 0, r.Width(), r.Height() + nBorder, pDestDC, r.left, r.top, SRCCOPY);
	
	//copy hRightDC -> hShadowRightDC
	hShadowRightDC.BitBlt(0, 0, r.Width(), r.Height(), &hRightDC, 0, 0, SRCCOPY);

	c = r.Width();
	i = c * nIntensity;

	for(x = 0; x < r.Width(); x++) 
	{
		for(y = 0; y < r.Height(); y++)
		{
			if(y <= c)
			{
				hShadowRightDC.SetPixel(x, y, 
					ColorAdjustLuma(hShadowRightDC.GetPixel(x, y), -in));

				in += nIntensity;
			}
			else if(y < r.Height() - c)

				hShadowRightDC.SetPixel(x, y, 
					ColorAdjustLuma(hShadowRightDC.GetPixel(x, y), -i));
		}

		c--;

		in = 0;
		i -= nIntensity;
	}

	pDestDC->BitBlt(r.left, r.top, r.Width(), r.Height(), &hShadowRightDC, 0, 0, SRCCOPY);
	
	// draw the bottom shadow

	r = GetBottomRect(rcSave);

	hBottomDC.BitBlt(0, 0, r.Width(), r.Height(), pDestDC, r.left, r.top, SRCCOPY);

	//copy hBottomDC -> hShadowBottomDC
	hShadowBottomDC.BitBlt(0, 0, r.Width(), r.Height(), &hBottomDC, 0, 0, SRCCOPY);
	
	c = r.Height();
	i = c * nIntensity;

	in = 0;
	
	for(y = 0; y < r.Height(); y++)
	{
		for(x = 0; x < r.Width(); x++) 
		{
			if(x <= c)
			{
				hShadowBottomDC.SetPixel(x, y, 
					ColorAdjustLuma(hShadowBottomDC.GetPixel(x, y), -in));

				in += nIntensity;
			}
			else if(x <= r.Width() - c)

				hShadowBottomDC.SetPixel(x, y, 
					ColorAdjustLuma(hShadowBottomDC.GetPixel(x, y), -i));
		}

		c--;

		in = 0;
		i -= nIntensity;
	}

	pDestDC->BitBlt(r.left, r.top, r.Width(), r.Height(), &hShadowBottomDC, 0, 0, SRCCOPY);

	// save the shadowed sections because for other possible uses such as menus

	hShadowBottomDC.BitBlt(0, 0, r.Width(), r.Height(), pDestDC, r.left, r.top, SRCCOPY);
	
	r = GetRightRect(rcSave);

	hShadowRightDC.BitBlt(0, 0, r.Width(), r.Height() + nBorder, pDestDC, r.left, r.top, SRCCOPY);
	
	// show that the shadow has been drawn at least once

	bDrawn = true;

	return true;
}

//////////////////////////////////////////////////////////////////////
// CRECT GETRIGHTRECT (CRECT R)
//////////////////////////////////////////////////////////////////////
//
//	GETS A RECTANGLE'S RIGHT SHADOW USING ANOTHER RECTANGLE AS
//  A POINT OF REFERENCE 
//
//////////////////////////////////////////////////////////////////////

CRect CShadow::GetRightRect(CRect r)
{
	r.left = r.right;
	r.right += nBorder;
	r.top += nBorder;
	r.bottom += nBorder;

	return r;
}

//////////////////////////////////////////////////////////////////////
// CRECT GETBOTTOMRECT (CRECT R)
//////////////////////////////////////////////////////////////////////
//
//	GETS A RECTANGLE'S BOTTOM SHADOW USING ANOTHER RECTANGLE AS
//  A POINT OF REFERENCE 
//
//////////////////////////////////////////////////////////////////////

CRect CShadow::GetBottomRect(CRect r)
{
	r.top = r.bottom;
	r.bottom += nBorder;
	r.left += nBorder;
	r.right += nBorder;

	return r;
}

COLORREF CShadow::ColorAdjustLuma(COLORREF hRef, int nIntensity)
{
	nIntensity >>= 2;

	int r = GetRValue(hRef) + nIntensity;
	int g = GetGValue(hRef) + nIntensity;
	int b = GetBValue(hRef) + nIntensity;

	if(r < 0) r = 0;
	if(g < 0) g = 0;
	if(b < 0) b = 0;

	return RGB(r, g, b);
}
